import yaml
import logging
from pymacaron_core.swagger.server import spawn_server_api
from pymacaron_core.swagger.client import generate_client_callers
from pymacaron_core.swagger.spec import ApiSpec
from pymacaron_core.models import get_model


log = logging.getLogger(__name__)


class APIClient():
    """Object gathering the API's client side code.
    Offers client-side methods to call every server endpoint declared in the API.
    """
    pass


class APIModels():
    """Object mapping constructors of API models to their names"""
    pass


def default_error_callback(e):
    """The default callback for handling exceptions caught in the client and server stubs:
    just raise the exception."""
    raise e


def generate_model_instantiator(model_name, definitions):
    # We need this to localize the value of model_class
    def instantiate_model(*args, **kwargs):
        return definitions.get(model_name)(*args, **kwargs)
    return instantiate_model


class API():
    """Describes a REST client/server API, with sugar coating:
    - easily instantiating the objects defined in the API
    - auto-generation of client code

    usage: See apipool.py
    """

    def __init__(self, name, yaml_str=None, yaml_path=None, timeout=10, error_callback=None, formats=None, do_persist=True, host=None, port=None, local=False, proto=None, verify_ssl=True):
        """An API Specification"""

        self.name = name

        # Is the endpoint callable directly as a python method from within the server?
        # (true is the flask server also serves that api)
        self.local = local

        # Callback to handle exceptions
        self.error_callback = default_error_callback

        # Flag: true if this api has spawned_api
        self.is_server = False
        self.app = None

        # Object holding the client side code to call the API
        self.client = APIClient()

        # Object holding constructors for the API's objects
        self.model = APIModels()

        self.client_timeout = timeout

        # Support versions of PyYAML with and without Loader
        import pkg_resources
        v = pkg_resources.get_distribution("PyYAML").version
        yamlkwargs = {}
        if v > '3.15':
            yamlkwargs['Loader'] = yaml.FullLoader

        if yaml_path:
            log.info("Loading swagger file at %s" % yaml_path)
            swagger_dict = yaml.load(open(yaml_path), **yamlkwargs)
        elif yaml_str:
            swagger_dict = yaml.load(yaml_str, **yamlkwargs)
        else:
            raise Exception("No swagger file specified")

        self.api_spec = ApiSpec(swagger_dict, formats, host, port, proto, verify_ssl)

        model_names = self.api_spec.load_models(do_persist=do_persist)

        # Add aliases to all models into self.model, so a developer may write:
        # 'ApiPool.<api_name>.model.<model_name>(*args)' to instantiate a model
        for model_name in model_names:
            setattr(self.model, model_name, get_model(model_name))

        if error_callback:
            self.error_callback = error_callback

        # Auto-generate client callers, so a developer may write:
        # 'ApiPool.<api_name>.call.login(param)' to call the login endpoint
        self._generate_client_callers()


    def _generate_client_callers(self, app=None):
        # If app is defined, we are doing local calls
        if app:
            callers_dict = generate_client_callers(self.api_spec, self.client_timeout, self.error_callback, True, app)
        else:
            callers_dict = generate_client_callers(self.api_spec, self.client_timeout, self.error_callback, False, None)

        for method, caller in list(callers_dict.items()):
            setattr(self.client, method, caller)


    def spawn_api(self, app, decorator=None):
        """Auto-generate server endpoints implementing the API into this Flask app"""
        if decorator:
            assert type(decorator).__name__ == 'function'
        self.is_server = True
        self.app = app

        if self.local:
            # Re-generate client callers, this time as local and passing them the app
            self._generate_client_callers(app)

        return spawn_server_api(self.name, app, self.api_spec, self.error_callback, decorator)


    def get_version(self):
        """Return the version of the API (as defined in the swagger file)"""
        return self.api_spec.version


    def model_to_json(self, object):
        """Take a model instance and return it as a json struct"""
        return object.to_json()


    def json_to_model(self, model_name, j, validate=False):
        """Take a json strust and a model name, and return a model instance"""
        if validate:
            self.api_spec.validate(model_name, j)
        o = getattr(self.model, model_name)
        return o.from_json(j)



import imp
import os
import pprint
import json
import logging

from flask import jsonify
from mock import patch

from pymacaron_core.models import get_model


utils = imp.load_source('common', os.path.join(os.path.dirname(__file__), 'utils.py'))


log = logging.getLogger(__name__)


class Test(utils.PymTest):


    def setUp(self):
        super(Test, self).setUp()


    def assertReplyOK(self, r, token):
        self.assertEqual(r.status_code, 200)
        j = json.loads(r.data.decode("utf-8"))
        print("json reply: " + pprint.pformat(j))
        self.assertEqual(j['token'], token)

    def assertError(self, r, status, content):
        self.assertEqual(str(r.status_code), str(status))
        self.assertTrue(content in r.status, "[%s] in [%s]" % (content, r.status))


    @patch('pymacaron_core.test.return_token')
    def test_swagger_server_no_param(self, func):
        func.__name__ = 'return_token'

        app, spec = self.generate_server_app(self.yaml_no_param)

        SessionToken = get_model('SessionToken')
        func.return_value = SessionToken(token='123')

        with app.test_client() as c:
            r = c.get('/v1/no/param')
            self.assertReplyOK(r, '123')
            func.assert_called_once_with()


    @patch('pymacaron_core.test.return_token')
    def test_swagger_server_no_result(self, func):
        func.__name__ = 'return_token'
        func.return_value = None

        app, spec = self.generate_server_app(self.yaml_no_param)

        with app.test_client() as c:
            r = c.get('/v1/no/param')
            self.assertError(r, 500, 'INTERNAL SERVER ERROR')


    @patch('pymacaron_core.test.return_token')
    def test_swagger_server_pass_through_responses(self, func):
        func.__name__ = 'return_token'

        app, spec = self.generate_server_app(self.yaml_no_param)

        with app.test_request_context('/'):
            r = jsonify({'foo': 'bar'})
            r.status_code = 534
            func.return_value = r

            with app.test_client() as c:
                r = c.get('/v1/no/param')
                print("sc: %s" % str(r.status_code))
                self.assertEqual(r.status_code, 534)
                j = json.loads(r.data.decode("utf-8"))
                print("json reply: " + pprint.pformat(j))
                self.assertEqual(j['foo'], 'bar')


    @patch('pymacaron_core.test.return_token')
    def test_swagger_invalid_server_return_value(self, func):
        func.__name__ = 'return_token'
        func.return_value = {'a': 1}

        app, spec = self.generate_server_app(self.yaml_no_param)

        with app.test_client() as c:
            r = c.get('/v1/no/param')
            self.assertError(r, 500, 'INTERNAL SERVER ERROR')


    @patch('pymacaron_core.test.return_token')
    def test_swagger_invalid_server_return_value_custom_callback(self, func):
        func.__name__ = 'return_token'
        func.return_value = {'a': 1}

        def callback(e):
            return get_model('SessionToken')(token=str(e))

        app, spec = self.generate_server_app(self.yaml_no_param, callback=callback)

        with app.test_client() as c:
            r = c.get('/v1/no/param')
            j = json.loads(r.data.decode('utf-8'))
            log.debug("Got reply: %s" % j)
            self.assertError(r, 500, 'INTERNAL SERVER ERROR')
            self.assertDictEqual(
                j,
                {'token': "Method pymacaron_core.test.return_token did not return a class instance but a <class 'dict'>"}
            )

# TODO: enable this test when server-side validation is enabled
#
#    c = get_model('Credentials')
#    func.return_value = c(email='asdasd')
#
#     with app.test_client() as c:
#         r = c.get('/v1/no/param')
#         print("sc: %s" % str(r.status_code))
#         assert r.status_code == 500, r.status_code
#         j = json.loads(r.data.decode("utf-8"))
#         print("json reply: " + pprint.pformat(j))
#         assert j['error'] == 'SERVER_ERROR'
#         assert 'did not return a class' in j['error_description']



    @patch('pymacaron_core.test.return_token')
    def test_swagger_server_param_in_body(self, func):
        func.__name__ = 'return_token'

        app, spec = self.generate_server_app(self.yaml_in_body)

        SessionToken = get_model('SessionToken')
        Credentials = get_model('Credentials')
        func.return_value = SessionToken(token='456')

        with app.test_client() as c:
            r = c.get('/v1/in/body', data=json.dumps({
                'email': 'a@a.a',
                'int': '123123',
            }))
            self.assertReplyOK(r, '456')
            func.assert_called_once_with(Credentials(email='a@a.a', int='123123'))


    @patch('pymacaron_core.test.return_token')
    def test_swagger_server_param_in_query(self, func):
        func.__name__ = 'return_token'

        app, spec = self.generate_server_app(self.yaml_in_query)

        SessionToken = get_model('SessionToken')
        func.return_value = SessionToken(token='456')

        with app.test_client() as c:
            r = c.get('/v1/in/query?foo=aaaa&bar=bbbb')
            self.assertReplyOK(r, '456')
            func.assert_called_once_with(bar='bbbb', foo='aaaa')


    @patch('pymacaron_core.test.return_token')
    def test_swagger_server_param_in_query__missing_required_param(self, func):
        func.__name__ = 'return_token'
        app, spec = self.generate_server_app(self.yaml_in_query)

        SessionToken = get_model('SessionToken')
        func.return_value = SessionToken(token='456')

        with app.test_client() as c:
            r = c.get('/v1/in/query?bar=bbbb')
            self.assertError(r, 400, 'BAD REQUEST')
            func.assert_not_called()


    def test_swagger_server_auth(self):
        pass


    @patch('pymacaron_core.test.return_token')
    def test_unmarshal_request_error__missing_required_argument(self, func):
        func.__name__ = 'return_token'

        app, spec = self.generate_server_app(self.yaml_in_body)

        SessionToken = get_model('SessionToken')
        func.return_value = SessionToken(token='456')

        with app.test_client() as c:
            r = c.get('/v1/in/body', data=json.dumps({'bazzoom': 'thiswontwork'}))
            self.assertError(r, 400, 'BAD REQUEST')
            func.assert_not_called()


    @patch('pymacaron_core.test.return_token')
    def test_unmarshal_request_error__wrong_argument_format(self, func):
        func.__name__ = 'return_token'

        app, spec = self.generate_server_app(self.yaml_in_body)

        SessionToken = get_model('SessionToken')
        func.return_value = SessionToken(token='456')

        with app.test_client() as c:
            data = json.dumps({
                'email': 'a@2a.a',
                'int': [1, 2, 3],
            })

            r = c.get('/v1/in/body', data=data)

        # verify server reply
# TODO: uncomment when bravado-core start returning error on this
#         assert r.status_code == 400, r.status_code
#         j = json.loads(r.data.decode("utf-8"))
#         print("json reply: " + pprint.pformat(j))
#         assert j['error'] == 'INVALID_PARAMETER'
#         assert j['status'] == 400
#         assert "email' is a required property" in j['error_description']



    @patch('pymacaron_core.test.return_token')
    def test_swagger_server_param_in_path(self, func):
        func.__name__ = 'return_token'

        app, spec = self.generate_server_app(self.yaml_in_path)

        SessionToken = get_model('SessionToken')
        func.return_value = SessionToken(token='456')

        with app.test_client() as c:
            r = c.get('/v1/in/1234/foo/bob234')
            self.assertReplyOK(r, '456')
            func.assert_called_once_with(item='1234', path='bob234')


# @patch('pymacaron_core.test.return_token')
# def test_swagger_server_param_in_path__missing_required_param(self, func):
#     app, spec = self.generate_server_app(self.yaml_in_query)

#     func.__name__ = 'return_token'
#     SessionToken = get_model('SessionToken')
#     func.return_value = SessionToken(token='456')

#     with app.test_client() as c:
#         r = c.get('/v1/in/query?bar=bbbb')
#         self.assertError(r, 400, 'BAD REQUEST')
#         func.assert_not_called()



def return_token(*args):
    # Mocked anyway
    pass

# Dummy implementation of model persistence
class PersistentFoo():

    @staticmethod
    def load_from_db():
        # Load object(s) from storage
        return 'bob'

    def save_to_db(object):
        # Put object into storage
        assert isinstance(object, PersistentFoo)
        return 'foo'


# Dummy implementation of model inheritance
class FunnyGrandad():

    def roflol(self):
        return 'roflol'

class FunnyDad(FunnyGrandad):

    def lol(self):
        return 'lol'
